/* Dia -- an diagram creation/manipulation program
 * Copyright (C) 1998 Alexander Larsson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>
#include <math.h>
#include <string.h>

#include <intl.h>
#include "object.h"
#include "connection.h"
#include "diarenderer.h"
#include "handle.h"
#include "arrows.h"
#include "properties.h"
/*#include "stereotype.h"*/
#include "attributes.h"

#include "pixmaps/relation.xpm"

#define PROP_RELATION_MAX_LINEWIDTH .5  /* width of a relation with a 100%
                                           strength*/

/* disturbances are represented by orthogonal lines to the main connector */
#define RELATION_MAX_DISTURBANCES 8     /* max number of disturbances */
#define RELATION_DISTURBANCE_WIDTH .1   /* line width */
#define RELATION_DISTURBANCE_LENGTH 1.5 /* length of each disturbance */
#define RELATION_DISTURBANCE_WOBBLE .4  /* amount of lateral displacement
                                           for multiple disturbances */
#define RELATION_DISTURBANCE_SPREAD 2.  /* distance between disturbances */

/* the flow characteristics of a relation are represented by small arrows
   pointing into the corresponding direction */
#define RELATION_MAX_FLOW 4             /* max number of arrows selectable */
#define RELATION_FLOW_WIDTH .1          /* line width */
#define RELATION_FLOW_LENGTH 1.5        /* length of each arrow */
#define RELATION_FLOW_SPREAD 1.         /* distance between arrows */
#define RELATION_FLOW_DISTANCE .5       /* distance to the relation line */
#define RELATION_FLOW_ARROWLEN .5       /* length of the arrow head */
#define RELATION_FLOW_ARROWWIDTH .5     /* width of the arrow head */

#define RELATION_LINEWIDTH(relation) ((relation)->strength / 100. * PROP_RELATION_MAX_LINEWIDTH)

typedef struct _Relation Relation;

struct _Relation {
  Connection connection;

  Handle feature_handle;

  Point feature_pos;
  int strength; /* how strong is this relationship? */

  int old_size;

  int disturbances;
  Point *dist_endpoints;

  int flows[2];
  Point *flow_endpoints[2];
  int flows_flip;

  Color line_color;
};

#define HANDLE_MOVE_FEATURE (HANDLE_CUSTOM1)

static ObjectChange* relation_move_handle(Relation *relation, Handle *handle,
                                            Point *to, ConnectionPoint *cp,
                                            HandleMoveReason reason, ModifierKeys modifiers);
static ObjectChange* relation_move(Relation *relation, Point *to);
static void relation_select(Relation *relation, Point *clicked_point,
                              DiaRenderer *interactive_renderer);
static void relation_draw(Relation *relation, DiaRenderer *renderer);
static DiaObject *relation_create(Point *startpoint,
                                 void *user_data,
                                 Handle **handle1,
                                 Handle **handle2);
static real relation_distance_from(Relation *relation, Point *point);
static void relation_update_data(Relation *relation);
static void relation_destroy(Relation *relation);

static PropDescription *relation_describe_props(Relation *relation);
static void relation_get_props(Relation * relation, GPtrArray *props);
static void relation_set_props(Relation * relation, GPtrArray *props);

static DiaObject *relation_load(ObjectNode obj_node, int version,
                               const char *filename);

static ObjectTypeOps relation_type_ops =
{
  (CreateFunc) relation_create,
  (LoadFunc)   relation_load,/*using_properties*/     /* load */
  (SaveFunc)   object_save_using_properties,      /* save */
  (GetDefaultsFunc)   NULL,
  (ApplyDefaultsFunc) NULL
};

DiaObjectType relation_type =
{
  "ECO - Relation",        /* name */
  0,                         /* version */
  (char **) relation_xpm,  /* pixmap */
  &relation_type_ops       /* ops */
};

static ObjectOps relation_ops = {
  (DestroyFunc)         relation_destroy,
  (DrawFunc)            relation_draw,
  (DistanceFunc)        relation_distance_from,
  (SelectFunc)          relation_select,
  (CopyFunc)            object_copy_using_properties,
  (MoveFunc)            relation_move,
  (MoveHandleFunc)      relation_move_handle,
  (GetPropertiesFunc)   object_create_props_dialog,
  (ApplyPropertiesFunc) object_apply_props_from_dialog,
  (ObjectMenuFunc)      NULL,
  (DescribePropsFunc)   relation_describe_props,
  (GetPropsFunc)        relation_get_props,
  (SetPropsFunc)        relation_set_props
};

static PropNumData prop_relation_strength_data = {1,100,5};
static PropNumData prop_relation_disturbances_data = {0,RELATION_MAX_DISTURBANCES,1};
static PropNumData prop_relation_flow_data = {0,4,1};


static PropDescription relation_props[] = {
  CONNECTION_COMMON_PROPERTIES,
  { "feature_pos", PROP_TYPE_POINT, 0, NULL, NULL, NULL},
  { "strength", PROP_TYPE_INT, PROP_FLAG_VISIBLE, N_("Strength:"), NULL, &prop_relation_strength_data},
  { "disturbances", PROP_TYPE_INT, PROP_FLAG_VISIBLE, N_("Disturbances:"), NULL, &prop_relation_disturbances_data},
  { "flow_forth", PROP_TYPE_INT, PROP_FLAG_VISIBLE, N_("Flow ->:"), NULL, &prop_relation_flow_data},
  { "flow_back", PROP_TYPE_INT, PROP_FLAG_VISIBLE, N_("Flow <-:"), NULL, &prop_relation_flow_data},
  { "flows_flip", PROP_TYPE_BOOL, PROP_FLAG_VISIBLE, N_("Switch flows"), NULL, NULL},
  /* can't use PROP_STD_TEXT_COLOUR_OPTIONAL cause it has PROP_FLAG_DONT_SAVE. It is designed to fill the Text object - not some subset */
  /*PROP_STD_TEXT_COLOUR_OPTIONS(PROP_FLAG_VISIBLE|PROP_FLAG_STANDARD|PROP_FLAG_OPTIONAL),*/
  PROP_STD_LINE_COLOUR_OPTIONAL,
  PROP_DESC_END
};

static PropDescription *
relation_describe_props(Relation *relation)
{
  if (relation_props[0].quark == 0) {
    prop_desc_list_calculate_quarks(relation_props);
  }
  return relation_props;
}

static PropOffset relation_offsets[] = {
  CONNECTION_COMMON_PROPERTIES_OFFSETS,
  { "feature_pos", PROP_TYPE_POINT, offsetof(Relation, feature_pos) },
  { "strength", PROP_TYPE_INT, offsetof(Relation, strength) },
  { "disturbances", PROP_TYPE_INT, offsetof(Relation, disturbances) },
  { "flow_forth", PROP_TYPE_INT, offsetof(Relation, flows[0]) },
  { "flow_back", PROP_TYPE_INT, offsetof(Relation, flows[1]) },
  { "flows_flip", PROP_TYPE_BOOL, offsetof(Relation, flows_flip) },
  { "line_colour",PROP_TYPE_COLOUR,offsetof(Relation, line_color)},
  { NULL, 0, 0 }
};

static void
relation_get_props(Relation * relation, GPtrArray *props)
{
  object_get_props_from_offsets(&relation->connection.object,
                                relation_offsets, props);
}

static void
relation_set_props(Relation *relation, GPtrArray *props)
{
  object_set_props_from_offsets(&relation->connection.object,
                                relation_offsets, props);
  relation_update_data(relation);
}

static real
relation_distance_from(Relation *relation, Point *point)
{
  Point *endpoints;
  real dist;

  endpoints = &relation->connection.endpoints[0];

  dist = distance_line_point(&endpoints[0], &endpoints[1],
                             RELATION_LINEWIDTH(relation), point);
  return dist;
}

static void
relation_select(Relation *relation, Point *clicked_point,
            DiaRenderer *interactive_renderer)
{
  connection_update_handles(&relation->connection);
}

static ObjectChange*
relation_move_handle(Relation *relation, Handle *handle,
                       Point *to, ConnectionPoint *cp,
                       HandleMoveReason reason, ModifierKeys modifiers)
{
  Point *endpoints;
  Point endpoints_delta, feature_delta;
  float dot;
  float length;

  assert(relation!=NULL);
  assert(handle!=NULL);
  assert(to!=NULL);

  endpoints = &relation->connection.endpoints[0];
  point_copy(&endpoints_delta, &endpoints[1]);
  point_sub(&endpoints_delta, &endpoints[0]);
  length = point_len(&endpoints_delta);
  if (handle->id == HANDLE_MOVE_FEATURE) {
    point_copy(&feature_delta, to);
    point_sub(&feature_delta, &endpoints[0]);
    point_copy(&relation->feature_pos, &endpoints[0]);
    dot = point_dot(&endpoints_delta, &feature_delta)/length/length;
    if(dot < 0) {
        dot = 0;
    } else if(dot > 1) {
        dot = 1;
    }
    point_add_scaled(&relation->feature_pos, &endpoints_delta, dot);
  } else  {
    point_copy(&feature_delta, &relation->feature_pos);
    point_sub(&feature_delta, &endpoints[0]);
    dot = point_dot(&endpoints_delta, &feature_delta)/length/length;
    if(!(dot > -.5 && dot < 1.5)) {
        dot = .5;
    }
    connection_move_handle(&relation->connection, handle->id, to, cp,
                           reason, modifiers);
    connection_adjust_for_autogap(&relation->connection);
    point_copy(&endpoints_delta, &endpoints[1]);
    point_sub(&endpoints_delta, &endpoints[0]);
    point_copy(&relation->feature_pos, &endpoints[0]);
    point_add_scaled(&relation->feature_pos, &endpoints_delta, dot);
  }

  relation_update_data(relation);

  return NULL;
}

static ObjectChange*
relation_move(Relation *relation, Point *to)
{
  Point start_to_end;
  Point *endpoints = &relation->connection.endpoints[0];
  Point delta;

  delta = *to;
  point_sub(&delta, &endpoints[0]);

  start_to_end = endpoints[1];
  point_sub(&start_to_end, &endpoints[0]);

  endpoints[1] = endpoints[0] = *to;
  point_add(&endpoints[1], &start_to_end);

  point_add(&relation->feature_pos, &delta);

  relation_update_data(relation);

  return NULL;
}

static void
relation_draw(Relation *relation, DiaRenderer *renderer)
{
  DiaRendererClass *renderer_ops = DIA_RENDERER_GET_CLASS (renderer);
  Point *endpoints;
  int i;
  Arrow arrow;

  assert(relation != NULL);
  assert(renderer != NULL);

  endpoints = &relation->connection.endpoints[0];

  renderer_ops->set_linewidth(renderer, RELATION_LINEWIDTH(relation));
  renderer_ops->set_linestyle(renderer, LINESTYLE_SOLID);
  renderer_ops->set_linecaps(renderer, LINECAPS_ROUND);

  renderer_ops->draw_line(renderer,
                                       &endpoints[0], &endpoints[1],
                                       &relation->line_color);

  renderer_ops->set_linewidth(renderer, RELATION_DISTURBANCE_WIDTH);
  for(i = 0; i < relation->disturbances * 2; i+=2) {
    renderer_ops->draw_line(renderer,
                                       &relation->dist_endpoints[i], &relation->dist_endpoints[i+1],
                                       &relation->line_color);
  }
  renderer_ops->set_linewidth(renderer, RELATION_FLOW_WIDTH);
  arrow.type = ARROW_FILLED_TRIANGLE;
  arrow.length = RELATION_FLOW_ARROWLEN;
  arrow.width = RELATION_FLOW_ARROWWIDTH;
  for(i = (relation->flows[0] + relation->flows[1]) * 2 - 2; i >= 0; i -= 2) {
    renderer_ops->draw_line_with_arrows(renderer,
                                       &relation->flow_endpoints[0][i], &relation->flow_endpoints[0][i+1],
                                       RELATION_FLOW_WIDTH,
                                       &relation->line_color,
                                       NULL, &arrow);
  }
}

static DiaObject *
relation_create(Point *startpoint,
                  void *user_data,
                  Handle **handle1,
                  Handle **handle2)
{
  Relation *relation;
  Connection *conn;
  DiaObject *obj;
  Point defaultlen = { 1.0, 1.0 };

  relation = g_malloc0(sizeof(Relation));

  conn = &relation->connection;
  conn->endpoints[0] = *startpoint;
  conn->endpoints[1] = *startpoint;
  point_add(&conn->endpoints[1], &defaultlen);

  obj = &conn->object;

  obj->type = &relation_type;
  obj->ops = &relation_ops;

  connection_init(conn, 3, 0);

  relation->strength = 50;
  relation->old_size = 0;
  relation->disturbances = 0;
  relation->dist_endpoints = NULL;

  relation->flows[0] = 0;
  relation->flows[1] = 0;
  relation->flow_endpoints[0] = 0;
  relation->flow_endpoints[1] = 0;

  relation->line_color = attributes_get_foreground();
  relation->feature_pos.x = 0.5*(conn->endpoints[0].x + conn->endpoints[1].x);
  relation->feature_pos.y = 0.5*(conn->endpoints[0].y + conn->endpoints[1].y);

  relation->feature_handle.id = HANDLE_MOVE_FEATURE;
  relation->feature_handle.type = HANDLE_MINOR_CONTROL;
  relation->feature_handle.connect_type = HANDLE_NONCONNECTABLE;
  relation->feature_handle.connected_to = NULL;
  obj->handles[2] = &relation->feature_handle;

  relation_update_data(relation);

  *handle1 = obj->handles[0];
  *handle2 = obj->handles[1];
  return &relation->connection.object;
}


static void
relation_destroy(Relation *relation)
{
  connection_destroy(&relation->connection);
  if(relation->dist_endpoints) {
    g_free(relation->dist_endpoints);
  }
}

static void
relation_add_disturbance(Relation *relation, int i, Point *pos, Point *normal, float scale) {
  point_copy(&relation->dist_endpoints[i], pos);
  point_copy(&relation->dist_endpoints[i+1], pos);
  point_add_scaled(&relation->dist_endpoints[i], normal, RELATION_DISTURBANCE_LENGTH/2.*scale);
  point_add_scaled(&relation->dist_endpoints[i+1], normal, RELATION_DISTURBANCE_LENGTH/-2.*scale);
}

static void
relation_update_data(Relation *relation)
{
  Connection *conn = &relation->connection;
  DiaObject *obj = &conn->object;
  LineBBExtras *extra;
  Rectangle rect;
  int i,dir;

  Point *endpoints;
  Point normal, straight; /* normalized diff/orthogonal vectors */

  int size; /* number of lines needed for all features */
  Point pos, *ppos;
  float sfactor, step, length;
  float scale; /* scale factors if features wouldn't fit on the line otherwise */

  size = relation->flows[0] +relation-> flows[1] +relation->disturbances;

  /* reallocate point array if necessary */
  if(relation->old_size != size) {
    if(relation->old_size) {
        g_free(relation->dist_endpoints);
        relation->dist_endpoints = NULL;
    }
    if(size > 0) {
        relation->dist_endpoints = g_malloc0(sizeof(Point)*2*size);
    }
    relation->old_size = size;
  }

  /* update flow pointers (they share a physical array with disturbances) */
  relation->flow_endpoints[0] = relation->dist_endpoints + 2*relation->disturbances;
  relation->flow_endpoints[1] = relation->flow_endpoints[0] + 2*relation->flows[0];

  /* some geometric info we only need if we have features */
  if(size > 0) {
    endpoints = &relation->connection.endpoints[0];
    normal.x = endpoints[1].y - endpoints[0].y;
    normal.y = endpoints[0].x - endpoints[1].x;
    length = point_len(&normal);
    point_scale(&normal, 1/length);
    point_normalize(&normal);
    straight.x = -normal.y;
    straight.y = normal.x;
  }

  /* calculate disturbance endpoints */
  if(relation->disturbances > 0) {
    if(length / 2 < RELATION_DISTURBANCE_SPREAD) {
        scale = length / 2 / RELATION_DISTURBANCE_SPREAD;
    } else {
        scale = 1;
    }
    if(relation->disturbances == 1) {
        relation_add_disturbance(relation, 0, &relation->feature_pos, &normal, scale);
    } else {
      sfactor = RELATION_DISTURBANCE_SPREAD/-2.;
      step = RELATION_DISTURBANCE_SPREAD/(relation->disturbances-1);
      for(i = 0; i < relation->disturbances * 2; i+=2, sfactor +=step) {
        point_copy(&pos, &relation->feature_pos);
        point_add_scaled(&pos, &straight, sfactor*scale);
        point_add_scaled(&pos, &normal, ((i/2+1)%3-1)*RELATION_DISTURBANCE_WOBBLE*scale);
        relation_add_disturbance(relation, i, &pos, &normal, scale);
      }
    }
  }

  /* calculate flow arrow endpoints */
  for(dir = 0; dir < 2; ++dir) {
    if(relation->flows[dir] > 0) {
      sfactor = -.5*RELATION_FLOW_LENGTH*relation->flows[dir] - .5*RELATION_FLOW_SPREAD*(relation->flows[dir] - 1);
      if(length < -2.2*sfactor) {
        scale = length/-2.2/sfactor;
      } else {
        scale = 1;
      }
      for(i = 0; i < relation->flows[dir]*2; ++i) {
        ppos = &relation->flow_endpoints[dir][i];
        point_copy(ppos, &relation->feature_pos);
        point_add_scaled(ppos, &straight, scale * sfactor * (dir*2 - 1));
        point_add_scaled(ppos, &normal, (RELATION_FLOW_DISTANCE + RELATION_LINEWIDTH(relation)/2.) * ((dir ^ relation->flows_flip)*2 - 1));
        sfactor += i % 2 ? RELATION_FLOW_SPREAD : RELATION_FLOW_LENGTH;
      }
    }
  }

  if (connpoint_is_autogap(conn->endpoint_handles[0].connected_to) ||
      connpoint_is_autogap(conn->endpoint_handles[1].connected_to)) {
    connection_adjust_for_autogap(conn);
  }
  obj->position = conn->endpoints[0];


  relation->feature_handle.pos = relation->feature_pos;

  connection_update_handles(conn);

  /* Boundingbox: */
  extra = &conn->extra_spacing;

  extra->start_long = extra->start_trans = extra->end_long = extra->end_trans = RELATION_LINEWIDTH(relation)/2.0;

  connection_update_boundingbox(conn);

  /* Extend bounding box for disturbances */
  extra->start_long = extra->start_trans = extra->end_long = extra->end_trans = RELATION_FLOW_ARROWWIDTH/2.0;
  for(i = 0; i < relation->disturbances; ++i) {
    line_bbox(&relation->dist_endpoints[i*2], &relation->dist_endpoints[i*2+1], extra, &rect);
    rectangle_union(&obj->bounding_box, &rect);
  }

  /* Extend bounding box for flow arrows */
  extra->start_long = extra->start_trans = extra->end_long = extra->end_trans = RELATION_DISTURBANCE_WIDTH/2.0;
  for(i = 0; i < 2; ++i) {
    if(!relation->flows[i]) {
      continue;
    }
    line_bbox(&relation->flow_endpoints[i][0], &relation->flow_endpoints[i][relation->flows[i]], extra, &rect);
    rectangle_union(&obj->bounding_box, &rect);
  }
}


static DiaObject *
relation_load(ObjectNode obj_node, int version, const char *filename)
{
  return object_load_using_properties(&relation_type,
                                      obj_node,version,filename);
}
